@@ -143,21 +143,18 @@ def check_follow_imports(choice: str) -> str:
 # types.
 ini_config_types: Final[dict[str, _INI_PARSER_CALLABLE]] = {
     "python_version": parse_version,
-    "strict_optional_whitelist": lambda s: s.split(),
     "custom_typing_module": str,
     "custom_typeshed_dir": expand_path,
     "mypy_path": lambda s: [expand_path(p.strip()) for p in re.split("[,:]", s)],
     "files": split_and_match_files,
     "quickstart_file": expand_path,
     "junit_xml": expand_path,
-    # These two are for backwards compatibility
-    "silent_imports": bool,
-    "almost_silent": bool,
     "follow_imports": check_follow_imports,
     "no_site_packages": bool,
     "plugins": lambda s: [p.strip() for p in s.split(",")],
     "always_true": lambda s: [p.strip() for p in s.split(",")],
     "always_false": lambda s: [p.strip() for p in s.split(",")],
+    "enable_incomplete_feature": lambda s: [p.strip() for p in s.split(",")],
     "disable_error_code": lambda s: validate_codes([p.strip() for p in s.split(",")]),
     "enable_error_code": lambda s: validate_codes([p.strip() for p in s.split(",")]),
     "package_root": lambda s: [p.strip() for p in s.split(",")],
@@ -165,6 +162,8 @@ ini_config_types: Final[dict[str, _INI_PARSER_CALLABLE]] = {
     "python_executable": expand_path,
     "strict": bool,
     "exclude": lambda s: [s.strip()],
+    "packages": try_split,
+    "modules": try_split,
 }
 
 # Reuse the ini_config_types and overwrite the diff
@@ -172,17 +171,19 @@ toml_config_types: Final[dict[str, _INI_PARSER_CALLABLE]] = ini_config_types.cop
 toml_config_types.update(
     {
         "python_version": parse_version,
-        "strict_optional_whitelist": try_split,
         "mypy_path": lambda s: [expand_path(p) for p in try_split(s, "[,:]")],
         "files": lambda s: split_and_match_files_list(try_split(s)),
         "follow_imports": lambda s: check_follow_imports(str(s)),
         "plugins": try_split,
         "always_true": try_split,
         "always_false": try_split,
+        "enable_incomplete_feature": try_split,
         "disable_error_code": lambda s: validate_codes(try_split(s)),
         "enable_error_code": lambda s: validate_codes(try_split(s)),
         "package_root": try_split,
         "exclude": str_or_array_as_list,
+        "packages": try_split,
+        "modules": try_split,
     }
 )
 
@@ -436,6 +437,9 @@ def parse_section(
                 elif key.startswith("disallow") and hasattr(template, key[3:]):
                     options_key = key[3:]
                     invert = True
+                elif key.startswith("show_") and hasattr(template, "hide_" + key[5:]):
+                    options_key = "hide_" + key[5:]
+                    invert = True
                 elif key == "strict":
                     pass  # Special handling below
                 else:
@@ -473,25 +477,6 @@ def parse_section(
             if v:
                 set_strict_flags()
             continue
-        if key == "silent_imports":
-            print(
-                "%ssilent_imports has been replaced by "
-                "ignore_missing_imports=True; follow_imports=skip" % prefix,
-                file=stderr,
-            )
-            if v:
-                if "ignore_missing_imports" not in results:
-                    results["ignore_missing_imports"] = True
-                if "follow_imports" not in results:
-                    results["follow_imports"] = "skip"
-        if key == "almost_silent":
-            print(
-                "%salmost_silent has been replaced by " "follow_imports=error" % prefix,
-                file=stderr,
-            )
-            if v:
-                if "follow_imports" not in results:
-                    results["follow_imports"] = "error"
         results[options_key] = v
 
     # These two flags act as per-module overrides, so store the empty defaults.
