@@ -31,32 +31,53 @@ Notes:
   even though some identities are preserved.
 """
 
-import contextlib
-from typing import Union, Iterator, Optional, Dict, Tuple
+from __future__ import annotations
+
+from contextlib import contextmanager, nullcontext
+from typing import Dict, Iterator, Tuple
+from typing_extensions import TypeAlias as _TypeAlias
 
 from mypy.nodes import (
-    FuncDef, NameExpr, MemberExpr, RefExpr, MypyFile, ClassDef, AssignmentStmt,
-    ImportFrom, CallExpr, Decorator, OverloadedFuncDef, Node, TupleExpr, ListExpr,
-    SuperExpr, IndexExpr, ImportAll, ForStmt, Block, CLASSDEF_NO_INFO, TypeInfo,
-    StarExpr, Var, SymbolTableNode
+    CLASSDEF_NO_INFO,
+    AssignmentStmt,
+    Block,
+    CallExpr,
+    ClassDef,
+    Decorator,
+    ForStmt,
+    FuncDef,
+    ImportAll,
+    ImportFrom,
+    IndexExpr,
+    ListExpr,
+    MemberExpr,
+    MypyFile,
+    NameExpr,
+    Node,
+    OverloadedFuncDef,
+    RefExpr,
+    StarExpr,
+    SuperExpr,
+    SymbolTableNode,
+    TupleExpr,
+    TypeInfo,
+    Var,
 )
 from mypy.traverser import TraverserVisitor
 from mypy.types import CallableType
 from mypy.typestate import TypeState
 
-
-SavedAttributes = Dict[Tuple[ClassDef, str], SymbolTableNode]
+SavedAttributes: _TypeAlias = Dict[Tuple[ClassDef, str], SymbolTableNode]
 
 
-def strip_target(node: Union[MypyFile, FuncDef, OverloadedFuncDef],
-                 saved_attrs: SavedAttributes) -> None:
-    """Reset a fine-grained incremental target to state before main pass of semantic analysis.
+def strip_target(
+    node: MypyFile | FuncDef | OverloadedFuncDef, saved_attrs: SavedAttributes
+) -> None:
+    """Reset a fine-grained incremental target to state before semantic analysis.
 
-    The most notable difference from the old version of strip_target() is that new semantic
-    analyzer doesn't have first pass where Vars and TypeInfos are pre-populated, so the stripping
-    is more thorough: all TypeInfos are killed. Therefore we need to preserve the variables
-    defined as attributes on self. This is done by patches (callbacks) returned from this function
-    that re-add these variables when called.
+    All TypeInfos are killed. Therefore we need to preserve the variables
+    defined as attributes on self. This is done by patches (callbacks)
+    returned from this function that re-add these variables when called.
 
     Args:
         node: node to strip
@@ -73,7 +94,7 @@ def strip_target(node: Union[MypyFile, FuncDef, OverloadedFuncDef],
 class NodeStripVisitor(TraverserVisitor):
     def __init__(self, saved_class_attrs: SavedAttributes) -> None:
         # The current active class.
-        self.type = None  # type: Optional[TypeInfo]
+        self.type: TypeInfo | None = None
         # This is True at class scope, but not in methods.
         self.is_class_body = False
         # By default, process function definitions. If False, don't -- this is used for
@@ -92,7 +113,7 @@ class NodeStripVisitor(TraverserVisitor):
         for name in file_node.names.copy():
             # TODO: this is a hot fix, we should delete all names,
             # see https://github.com/python/mypy/issues/6422.
-            if '@' not in name:
+            if "@" not in name:
                 del file_node.names[name]
 
     def visit_block(self, b: Block) -> None:
@@ -114,13 +135,15 @@ class NodeStripVisitor(TraverserVisitor):
         node.type_vars = []
         node.base_type_exprs.extend(node.removed_base_type_exprs)
         node.removed_base_type_exprs = []
-        node.defs.body = [s for s in node.defs.body
-                          if s not in to_delete]  # type: ignore[comparison-overlap]
+        node.defs.body = [
+            s for s in node.defs.body if s not in to_delete  # type: ignore[comparison-overlap]
+        ]
         with self.enter_class(node.info):
             super().visit_class_def(node)
         TypeState.reset_subtype_caches_for(node.info)
         # Kill the TypeInfo, since there is none before semantic analysis.
         node.info = CLASSDEF_NO_INFO
+        node.analyzed = None
 
     def save_implicit_attributes(self, node: ClassDef) -> None:
         """Produce callbacks that re-add attributes defined on self."""
@@ -140,7 +163,7 @@ class NodeStripVisitor(TraverserVisitor):
             # See also #4814.
             assert isinstance(node.type, CallableType)
             node.type.variables = []
-        with self.enter_method(node.info) if node.info else nothing():
+        with self.enter_method(node.info) if node.info else nullcontext():
             super().visit_func_def(node)
 
     def visit_decorator(self, node: Decorator) -> None:
@@ -230,7 +253,7 @@ class NodeStripVisitor(TraverserVisitor):
         elif isinstance(lvalue, StarExpr):
             self.process_lvalue_in_method(lvalue.expr)
 
-    @contextlib.contextmanager
+    @contextmanager
     def enter_class(self, info: TypeInfo) -> Iterator[None]:
         old_type = self.type
         old_is_class_body = self.is_class_body
@@ -240,7 +263,7 @@ class NodeStripVisitor(TraverserVisitor):
         self.type = old_type
         self.is_class_body = old_is_class_body
 
-    @contextlib.contextmanager
+    @contextmanager
     def enter_method(self, info: TypeInfo) -> Iterator[None]:
         old_type = self.type
         old_is_class_body = self.is_class_body
@@ -249,8 +272,3 @@ class NodeStripVisitor(TraverserVisitor):
         yield
         self.type = old_type
         self.is_class_body = old_is_class_body
-
-
-@contextlib.contextmanager
-def nothing() -> Iterator[None]:
-    yield
