@@ -209,7 +209,7 @@ class TypeTranslator(TypeVisitor[Type]):
         last_known_value: LiteralType | None = None
         if t.last_known_value is not None:
             raw_last_known_value = t.last_known_value.accept(self)
-            assert isinstance(raw_last_known_value, LiteralType)  # type: ignore
+            assert isinstance(raw_last_known_value, LiteralType)  # type: ignore[misc]
             last_known_value = raw_last_known_value
         return Instance(
             typ=t.type,
@@ -266,7 +266,7 @@ class TypeTranslator(TypeVisitor[Type]):
 
     def visit_literal_type(self, t: LiteralType) -> Type:
         fallback = t.fallback.accept(self)
-        assert isinstance(fallback, Instance)  # type: ignore
+        assert isinstance(fallback, Instance)  # type: ignore[misc]
         return LiteralType(value=t.value, fallback=fallback, line=t.line, column=t.column)
 
     def visit_union_type(self, t: UnionType) -> Type:
@@ -284,7 +284,7 @@ class TypeTranslator(TypeVisitor[Type]):
         items: list[CallableType] = []
         for item in t.items:
             new = item.accept(self)
-            assert isinstance(new, CallableType)  # type: ignore
+            assert isinstance(new, CallableType)  # type: ignore[misc]
             items.append(new)
         return Overloaded(items=items)
 
