@@ -9,14 +9,31 @@ We need to split basic blocks on each error check since branches can
 only be placed at the end of a basic block.
 """
 
-from typing import List, Optional
+from __future__ import annotations
 
-from mypyc.ir.ops import (
-    Value, BasicBlock, LoadErrorValue, Return, Branch, RegisterOp, Integer, ERR_NEVER, ERR_MAGIC,
-    ERR_FALSE, ERR_ALWAYS, NO_TRACEBACK_LINE_NO
-)
 from mypyc.ir.func_ir import FuncIR
+from mypyc.ir.ops import (
+    ERR_ALWAYS,
+    ERR_FALSE,
+    ERR_MAGIC,
+    ERR_MAGIC_OVERLAPPING,
+    ERR_NEVER,
+    NO_TRACEBACK_LINE_NO,
+    BasicBlock,
+    Branch,
+    CallC,
+    ComparisonOp,
+    GetAttr,
+    Integer,
+    LoadErrorValue,
+    RegisterOp,
+    Return,
+    SetAttr,
+    Value,
+)
 from mypyc.ir.rtypes import bool_rprimitive
+from mypyc.primitives.exc_ops import err_occurred_op
+from mypyc.primitives.registry import CFunctionDescription
 
 
 def insert_exception_handling(ir: FuncIR) -> None:
@@ -25,6 +42,7 @@ def insert_exception_handling(ir: FuncIR) -> None:
     # block. The block just returns an error value.
     error_label = None
     for block in ir.blocks:
+        adjust_error_kinds(block)
         can_raise = any(op.can_raise() for op in block.ops)
         if can_raise:
             error_label = add_handler_block(ir)
@@ -42,10 +60,10 @@ def add_handler_block(ir: FuncIR) -> BasicBlock:
     return block
 
 
-def split_blocks_at_errors(blocks: List[BasicBlock],
-                           default_error_handler: BasicBlock,
-                           func_name: Optional[str]) -> List[BasicBlock]:
-    new_blocks: List[BasicBlock] = []
+def split_blocks_at_errors(
+    blocks: list[BasicBlock], default_error_handler: BasicBlock, func_name: str | None
+) -> list[BasicBlock]:
+    new_blocks: list[BasicBlock] = []
 
     # First split blocks on ops that may raise.
     for block in blocks:
@@ -81,19 +99,36 @@ def split_blocks_at_errors(blocks: List[BasicBlock],
                     # this is a hack to represent the always fail
                     # semantics, using a temporary bool with value false
                     target = Integer(0, bool_rprimitive)
+                elif op.error_kind == ERR_MAGIC_OVERLAPPING:
+                    errvalue = Integer(int(target.type.c_undefined), rtype=op.type)
+                    comp = ComparisonOp(target, errvalue, ComparisonOp.EQ)
+                    cur_block.ops.append(comp)
+                    new_block2 = BasicBlock()
+                    new_blocks.append(new_block2)
+                    branch = Branch(
+                        comp,
+                        true_label=new_block2,
+                        false_label=new_block,
+                        op=Branch.BOOL,
+                        rare=True,
+                    )
+                    cur_block.ops.append(branch)
+                    cur_block = new_block2
+                    target = primitive_call(err_occurred_op, [], target.line)
+                    cur_block.ops.append(target)
+                    variant = Branch.IS_ERROR
+                    negated = True
                 else:
-                    assert False, 'unknown error kind %d' % op.error_kind
+                    assert False, "unknown error kind %d" % op.error_kind
 
                 # Void ops can't generate errors since error is always
                 # indicated by a special value stored in a register.
                 if op.error_kind != ERR_ALWAYS:
                     assert not op.is_void, "void op generating errors?"
 
-                branch = Branch(target,
-                                true_label=error_label,
-                                false_label=new_block,
-                                op=variant,
-                                line=op.line)
+                branch = Branch(
+                    target, true_label=error_label, false_label=new_block, op=variant, line=op.line
+                )
                 branch.negated = negated
                 if op.line != NO_TRACEBACK_LINE_NO and func_name is not None:
                     branch.traceback_entry = (func_name, op.line)
@@ -101,3 +136,30 @@ def split_blocks_at_errors(blocks: List[BasicBlock],
                 cur_block = new_block
 
     return new_blocks
+
+
+def primitive_call(desc: CFunctionDescription, args: list[Value], line: int) -> CallC:
+    return CallC(
+        desc.c_function_name,
+        [],
+        desc.return_type,
+        desc.steals,
+        desc.is_borrowed,
+        desc.error_kind,
+        line,
+    )
+
+
+def adjust_error_kinds(block: BasicBlock) -> None:
+    """Infer more precise error_kind attributes for ops.
+
+    We have access here to more information than what was available
+    when the IR was initially built.
+    """
+    for op in block.ops:
+        if isinstance(op, GetAttr):
+            if op.class_type.class_ir.is_always_defined(op.attr):
+                op.error_kind = ERR_NEVER
+        if isinstance(op, SetAttr):
+            if op.class_type.class_ir.is_always_defined(op.attr):
+                op.error_kind = ERR_NEVER
